# Copyright © 2016–2023 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Parser for XPath 1.0 expressions.

This module defines the following classes:

* XPathParser: Recursive descent parser for XPath expressions.

The module also defines the following exceptions:

"""
from typing import Optional, Union
from .schemadata import SchemaContext
from .enumerations import Axis, MultiplicativeOp
from .exceptions import EndOfInput, InvalidXPath, NotSupported, UnexpectedInput
from .parser import Parser
from .typealiases import QualName
from .xpathast import (
    AdditiveExpr, AndExpr, EqualityExpr, Expr, FilterExpr, FuncBitIsSet,
    FuncBoolean, FuncCeiling, FuncConcat, FuncContains, FuncCount, FuncCurrent,
    FuncDeref, FuncDerivedFrom, FuncEnumValue, FuncFalse, FuncFloor, FuncLast,
    FuncName, FuncNormalizeSpace, FuncNot, FuncNumber, FuncPosition,
    FuncReMatch, FuncRound, FuncStartsWith, FuncString, FuncStringLength,
    FuncSubstring, FuncSubstringAfter, FuncSubstringBefore, FuncSum,
    FuncTranslate, FuncTrue, Literal, LocationPath, MultiplicativeExpr, Number,
    OrExpr, PathExpr, RelationalExpr, Root, Step, UnaryMinusExpr, UnionExpr)


class XPathParser(Parser):
    """Parser for XPath expressions."""

    def __init__(self: "XPathParser", text: str, sctx: SchemaContext):
        """Initialize the parser instance.

        Args:
            sctx: Schema context for XPath expression parsing.
        """
        super().__init__(text)
        self.sctx = sctx

    def parse(self: "XPathParser") -> Expr:
        """Parse an XPath expression.

        Returns:
            A node of the XPath AST.

        Raises:
            InvalidXPath: If the input XPath expression is invalid.
            NotSupported: If the input XPath expression contains a feature
                that isn't supported by the implementation.
        """
        self.skip_ws()
        return self._or_expr()

    def _or_expr(self: "XPathParser") -> Expr:
        op1 = self._and_expr()
        while self.test_string("or"):
            self.skip_ws()
            op2 = self._and_expr()
            op1 = OrExpr(op1, op2)
        return op1

    def _and_expr(self: "XPathParser") -> Expr:
        op1 = self._equality_expr()
        while self.test_string("and"):
            self.skip_ws()
            op2 = self._equality_expr()
            op1 = AndExpr(op1, op2)
        return op1

    def _equality_expr(self: "XPathParser") -> Expr:
        op1 = self._relational_expr()
        while True:
            negate = False
            try:
                next = self.peek()
            except EndOfInput:
                return op1
            if next == "!":
                self.offset += 1
                negate = True
                try:
                    next = self.peek()
                except EndOfInput:
                    raise InvalidXPath(self)
            if next != "=":
                if negate:
                    raise InvalidXPath(self)
                return op1
            self.adv_skip_ws()
            op2 = self._relational_expr()
            op1 = EqualityExpr(op1, op2, negate)

    def _relational_expr(self: "XPathParser") -> Expr:
        op1 = self._additive_expr()
        while True:
            try:
                rel = self.peek()
            except EndOfInput:
                return op1
            if rel not in "<>":
                return op1
            self.offset += 1
            eq = self.test_string("=")
            self.skip_ws()
            op2 = self._additive_expr()
            op1 = RelationalExpr(op1, op2, rel == "<", eq)

    def _additive_expr(self: "XPathParser") -> Expr:
        op1 = self._multiplicative_expr()
        while True:
            try:
                pm = self.peek()
            except EndOfInput:
                return op1
            if pm not in "+-":
                return op1
            self.adv_skip_ws()
            op2 = self._multiplicative_expr()
            op1 = AdditiveExpr(op1, op2, pm == "+")

    def _multiplicative_expr(self: "XPathParser") -> Expr:
        op1 = self._unary_minus_expr()
        while True:
            if self.test_string("*"):
                mulop = MultiplicativeOp.multiply
            elif self.test_string("div"):
                mulop = MultiplicativeOp.divide
            elif self.test_string("mod"):
                mulop = MultiplicativeOp.modulo
            else:
                return op1
            self.skip_ws()
            op2 = self._unary_minus_expr()
            op1 = MultiplicativeExpr(op1, op2, mulop)

    def _unary_minus_expr(self: "XPathParser") -> Expr:
        negate = None
        while self.test_string("-"):
            negate = not negate
            self.skip_ws()
        expr = self._union_expr()
        return expr if negate is None else UnaryMinusExpr(expr, negate)

    def _union_expr(self: "XPathParser") -> Expr:
        op1 = self._lit_num_path()
        while self.test_string("|"):
            self.skip_ws()
            op2 = self._lit_num_path()
            op1 = UnionExpr(op1, op2)
        return op1

    def _lit_num_path(self: "XPathParser") -> Expr:
        next = self.peek()
        if next == "(":
            self.adv_skip_ws()
            return self._path_expr(None)
        if next in "'\"":
            self.offset += 1
            val = self.up_to(next)
            self.skip_ws()
            return Literal(val)
        if ("0" <= next <= "9" or
                next == "." and not self.at_last_char() and
                "0" <= self.input[self.offset + 1] <= "9"):
            val = self.unsigned_float()
            self.skip_ws()
            return Number(val)
        start = self.offset
        try:
            fname = self.yang_identifier()
        except UnexpectedInput:
            return self._location_path()
        self.skip_ws()
        if self.test_string("(") and fname not in (
                "node", "comment", "processing-instruction", "text"):
            self.skip_ws()
            return self._path_expr(fname)
        self.offset = start
        return self._location_path()

    def _path_expr(self: "XPathParser", fname: str) -> Expr:
        fexpr = self._filter_expr(fname)
        if self.test_string("/"):
            return PathExpr(fexpr, self._location_path())
        return fexpr

    def _filter_expr(self: "XPathParser", fname: str) -> Expr:
        if fname is None:
            prim = self._or_expr()
        else:
            mname = "_func_" + fname.replace("-", "_")
            try:
                prim = getattr(self, mname)()
            except AttributeError:
                if fname in ("id", "lang", "namespace-uri"):
                    raise NotSupported(self, f"function '{fname}()'") from None
                raise InvalidXPath(self) from None
        self.char(")")
        self.skip_ws()
        return FilterExpr(prim, self._predicates())

    def _predicates(self: "XPathParser") -> list[Expr]:
        res = []
        while self.test_string("["):
            self.skip_ws()
            res.append(self._predicate())
            self.char("]")
            self.skip_ws()
        return res

    def _predicate(self: "XPathParser") -> Expr:
        return self._or_expr()

    def _location_path(self: "XPathParser") -> LocationPath:
        if self.test_string("/"):
            self.skip_ws()
            if self.at_end():
                return Root()
            op1 = LocationPath(Root(), self._step())
        else:
            op1 = self._step()
        while self.test_string("/"):
            self.skip_ws()
            op2 = self._step()
            op1 = LocationPath(op1, op2)
        return op1

    def _step(self: "XPathParser") -> Step:
        return Step(*self._axis_qname(), self._predicates())

    def _axis_qname(self: "XPathParser") -> tuple[Axis, Union[QualName, bool, None]]:
        next = self.peek()
        if next == "*":
            self.adv_skip_ws()
            return (Axis.child, False)
        if next == "/":
            self.skip_ws()
            return (Axis.descendant_or_self, None)
        if next == ".":
            self.offset += 1
            res = (Axis.parent if self.test_string(".") else Axis.self, None)
            self.skip_ws()
            return res
        try:
            yid = self.yang_identifier()
        except UnexpectedInput:
            raise InvalidXPath(self) from None
        ws = self.skip_ws()
        try:
            next = self.peek()
        except EndOfInput:
            return Axis.child, (yid, self.sctx.default_ns)
        if next == "(":
            return (Axis.child, self._node_type(yid))
        if next == ":":
            self.offset += 1
            next = self.peek()
            if next == ":":
                self.adv_skip_ws()
                try:
                    axis = Axis[yid.replace("-", "_")]
                except KeyError:
                    if yid in ("attribute", "following",
                               "namespace", "preceding"):
                        raise NotSupported(self, f"axis '{yid}::'") from None
                    raise InvalidXPath(self) from None
                return axis, self._qname()
            if ws:
                raise InvalidXPath(self)
            nsp = self.sctx.schema_data.prefix2ns(yid, self.sctx.text_mid)
            loc = self.yang_identifier()
            self.skip_ws()
            return Axis.child, (loc, nsp)
        return Axis.child, (yid, self.sctx.default_ns)

    def _node_type(self: "XPathParser", typ):
        if typ == "node":
            self.adv_skip_ws()
            self.char(")")
            self.skip_ws()
            return None
        elif typ in ("comment", "processing-instruction", "text"):
            raise NotSupported(self, f"node type '{typ}()'")
        raise InvalidXPath(self)

    def _qname(self: "XPathParser") -> Optional[QualName]:
        """Parse XML QName."""
        if self.test_string("*"):
            self.skip_ws()
            return False
        ident = self.yang_identifier()
        ws = self.skip_ws()
        try:
            next = self.peek()
        except EndOfInput:
            return ident, self.sctx.default_ns
        if next == "(":
            return self._node_type(ident)
        if not ws and self.test_string(":"):
            res = (
                self.yang_identifier(),
                self.sctx.schema_data.prefix2ns(ident, self.sctx.text_mid))
        else:
            res = (ident, self.sctx.default_ns)
        self.skip_ws()
        return res

    def _opt_arg(self: "XPathParser") -> Optional[Expr]:
        return None if self.peek() == ")" else self.parse()

    def _two_args(self: "XPathParser") -> tuple[Expr, Expr]:
        fst = self.parse()
        self.char(",")
        self.skip_ws()
        return fst, self.parse()

    def _func_bit_is_set(self: "XPathParser") -> FuncBitIsSet:
        return FuncBitIsSet(*self._two_args())

    def _func_boolean(self: "XPathParser") -> FuncBoolean:
        return FuncBoolean(self.parse())

    def _func_ceiling(self: "XPathParser") -> FuncCeiling:
        return FuncCeiling(self.parse())

    def _func_concat(self: "XPathParser") -> FuncConcat:
        res = [self.parse()]
        while self.test_string(","):
            self.skip_ws()
            res.append(self.parse())
        if len(res) < 2:
            raise InvalidXPath(self)
        return FuncConcat(res)

    def _func_contains(self: "XPathParser") -> FuncContains:
        return FuncContains(*self._two_args())

    def _func_count(self: "XPathParser") -> FuncCount:
        return FuncCount(self.parse())

    def _func_current(self: "XPathParser") -> FuncCurrent:
        return FuncCurrent()

    def _func_deref(self: "XPathParser") -> FuncDeref:
        return FuncDeref(self.parse())

    def _func_derived_from(self: "XPathParser") -> FuncDerivedFrom:
        return FuncDerivedFrom(*self._two_args(), False, self.sctx)

    def _func_derived_from_or_self(self: "XPathParser") -> FuncDerivedFrom:
        return FuncDerivedFrom(*self._two_args(), True, self.sctx)

    def _func_enum_value(self: "XPathParser") -> FuncEnumValue:
        return FuncEnumValue(self.parse())

    def _func_false(self: "XPathParser") -> FuncFalse:
        return FuncFalse()

    def _func_floor(self: "XPathParser") -> FuncFloor:
        return FuncFloor(self.parse())

    def _func_last(self: "XPathParser") -> FuncLast:
        return FuncLast()

    def _func_local_name(self: "XPathParser") -> FuncName:
        return FuncName(self._opt_arg(), local=True)

    def _func_name(self: "XPathParser") -> FuncName:
        return FuncName(self._opt_arg(), local=False)

    def _func_normalize_space(self: "XPathParser") -> FuncNormalizeSpace:
        return FuncNormalizeSpace(self._opt_arg())

    def _func_not(self: "XPathParser") -> FuncNot:
        return FuncNot(self.parse())

    def _func_number(self: "XPathParser") -> FuncNumber:
        return FuncNumber(self._opt_arg())

    def _func_position(self: "XPathParser") -> FuncPosition:
        return FuncPosition()

    def _func_re_match(self: "XPathParser") -> FuncReMatch:
        return FuncReMatch(*self._two_args())

    def _func_round(self: "XPathParser") -> FuncRound:
        return FuncRound(self.parse())

    def _func_starts_with(self: "XPathParser") -> FuncStartsWith:
        return FuncStartsWith(*self._two_args())

    def _func_string(self: "XPathParser") -> FuncString:
        return FuncString(self._opt_arg())

    def _func_string_length(self: "XPathParser") -> FuncStringLength:
        return FuncStringLength(self._opt_arg())

    def _func_substring(self: "XPathParser") -> FuncSubstring:
        string, start = self._two_args()
        if self.test_string(","):
            self.skip_ws()
            length = self.parse()
        else:
            length = None
        return FuncSubstring(string, start, length)

    def _func_substring_after(self: "XPathParser") -> FuncSubstringAfter:
        return FuncSubstringAfter(*self._two_args())

    def _func_substring_before(self: "XPathParser") -> FuncSubstringBefore:
        return FuncSubstringBefore(*self._two_args())

    def _func_sum(self: "XPathParser") -> FuncSum:
        return FuncSum(self.parse())

    def _func_translate(self: "XPathParser") -> FuncTranslate:
        s1, s2 = self._two_args()
        self.char(",")
        self.skip_ws()
        return FuncTranslate(s1, s2, self.parse())

    def _func_true(self: "XPathParser") -> FuncTrue:
        return FuncTrue()



# Copyright © 2016–2023 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""This module defines the entry point for a validation script."""

import argparse
import json
import os
import sys
import pkg_resources
from yangson import DataModel
from yangson.enumerations import ContentType, ValidationScope
from yangson.exceptions import (
    BadYangLibraryData, FeaturePrerequisiteError, MultipleImplementedRevisions,
    ModuleNotFound, ModuleNotRegistered, RawMemberError, RawTypeError,
    SchemaError, SemanticError, YangTypeError)


def main(ylib: str = None, path: str = None,
         scope: ValidationScope = ValidationScope.all,
         ctype: ContentType = ContentType.config, set_id: bool = False,
         tree: bool = False, no_types: bool = False,
         digest: bool = False, validate: str = None) -> int:
    """Entry-point for a validation script.

    Args:
        ylib: Name of the file with YANG library
        path: Colon-separated list of directories to search  for YANG modules.
        scope: Validation scope (syntax, semantics or all).
        ctype: Content type of the data instance (config, nonconfig or all)
        set_id: If `True`, print module set id.
        tree: If `True`, print schema tree.
        no_types: If `True`, don't print types in schema tree.
        digest: If `True`, print schema digest.
        validate: Name of file to validate against the schema.

    Returns:
        Numeric return code (0=no error, 2=YANG error, 1=other)
    """
    if ylib is None:
        parser = argparse.ArgumentParser(
            prog="yangson",
            description="Validate JSON data against a YANG data model.")
        parser.add_argument(
            "-V", "--version", action="version",
            version=f"%(prog)s {pkg_resources.get_distribution('yangson').version}")
        parser.add_argument(
            "ylib", metavar="YLIB",
            help=("name of the file with description of the data model"
                  " in JSON-encoded YANG library format [RFC 7895]"))
        parser.add_argument(
            "-p", "--path",
            help=("colon-separated list of directories to search"
                  " for YANG modules"))
        grp = parser.add_mutually_exclusive_group()
        grp.add_argument(
            "-i", "--id", action="store_true",
            help="print module set id")
        grp.add_argument(
            "-t", "--tree", action="store_true",
            help="print schema tree as ASCII art")
        grp.add_argument(
            "-d", "--digest", action="store_true",
            help="print schema digest in JSON format")
        grp.add_argument(
            "-v", "--validate", metavar="INST",
            help="name of the file with JSON-encoded instance data")
        parser.add_argument(
            "-s", "--scope", choices=["syntax", "semantics", "all"],
            default="all", help="validation scope (default: %(default)s)")
        parser.add_argument(
            "-c", "--ctype", type=str, choices=["config", "nonconfig", "all"],
            default="config",
            help="content type of the data instance (default: %(default)s)")
        parser.add_argument(
            "-n", "--no-types", action="store_true",
            help="suppress type info in tree output")
        args = parser.parse_args()
        ylib: str = args.ylib
        path: Optional[str] = args.path
        scope = ValidationScope[args.scope]
        ctype = ContentType[args.ctype]
        set_id: bool = args.id
        tree: bool = args.tree
        no_types = args.no_types
        digest: bool = args.digest
        validate: str = args.validate
    try:
        with open(ylib, encoding="utf-8") as infile:
            yl = infile.read()
    except (FileNotFoundError, PermissionError,
            json.decoder.JSONDecodeError) as e:
        print("YANG library:", str(e), file=sys.stderr)
        return 1
    sp = path if path else os.environ.get("YANG_MODPATH", ".")
    try:
        dm = DataModel(yl, tuple(sp.split(":")))
    except BadYangLibraryData as e:
        print("Invalid YANG library:", str(e), file=sys.stderr)
        return 2
    except FeaturePrerequisiteError as e:
        print("Unsupported pre-requisite feature:", str(e), file=sys.stderr)
        return 2
    except MultipleImplementedRevisions as e:
        print("Multiple implemented revisions:", str(e), file=sys.stderr)
        return 2
    except ModuleNotFound as e:
        print("Module not found:", str(e), file=sys.stderr)
        return 2
    except ModuleNotRegistered as e:
        print("Module not registered:", str(e), file=sys.stderr)
        return 2
    if set_id:
        print(dm.module_set_id())
        return 0
    if tree:
        print(dm.ascii_tree(no_types))
        return 0
    if digest:
        print(dm.schema_digest())
        return 0
    if not validate:
        return 0
    try:
        with open(validate, encoding="utf-8") as infile:
            itxt = json.load(infile)
    except (FileNotFoundError, PermissionError,
            json.decoder.JSONDecodeError) as e:
        print("Instance data:", str(e), file=sys.stderr)
        return 1
    try:
        i = dm.from_raw(itxt)
    except RawMemberError as e:
        print("Illegal object member:", str(e), file=sys.stderr)
        return 3
    except RawTypeError as e:
        print("Invalid type:", str(e), file=sys.stderr)
        return 3
    try:
        i.validate(scope, ctype)
    except SchemaError as e:
        print("Schema error:", str(e), file=sys.stderr)
        return 3
    except SemanticError as e:
        print("Semantic error:", str(e), file=sys.stderr)
        return 3
    except YangTypeError as e:
        print("Invalid type:", str(e), file=sys.stderr)
        return 3
    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright © 2016–2023 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Classes representing YANG data types.

This module implements the following classes:

* BitsType: YANG bits type.
* BinaryType: YANG binary type.
* BooleanType: YANG boolean type.
* DataType: Abstract class for data types.
* Decimal64Type: YANG decimal64 type.
* EmptyType: YANG empty type.
* EnumerationType: YANG enumeration type.
* IdentityrefType: YANG identityref type.
* InstanceIdentifierType: YANG instance-identifier type.
* LeafrefType: YANG leafref type.
* LinkType: Abstract class for data types representing links.
* IntegralType: Abstract class for integral types.
* Int8Type: YANG int8 type.
* Int16Type: YANG int16 type.
* Int32Type: YANG int32 type.
* Int64Type: YANG int64 type.
* NumericType: Abstract class for numeric types.
* StringType: YANG string type.
* Uint8Type: YANG uint8 type.
* Uint16Type: YANG uint16 type.
* Uint32Type: YANG uint32 type.
* Uint64Type: YANG uint64 type.
* UnionType: YANG union type.
"""
import base64
import binascii
import decimal
import numbers
import xml.etree.ElementTree as ET
import re
from typing import Any, Optional, Union, TYPE_CHECKING

from .constraint import Intervals, Pattern
from .exceptions import (
    InvalidArgument, ParserException, ModuleNotRegistered, UnknownPrefix,
    InvalidLeafrefPath, MissingModuleNamespace)
from .schemadata import SchemaContext
from .instance import InstanceNode, InstanceIdParser, InstanceRoute
from .statement import Statement
from .typealiases import QualName, RawScalar, ScalarValue, YangIdentifier
from .xpathparser import XPathParser
if TYPE_CHECKING:
    from .schemanode import TerminalNode


class DataType:
    """Abstract class for YANG data types."""

    _option_template = '<option value="{}"{}>{}</option>'

    def __init__(self: "DataType", sctx: SchemaContext, name: Optional[YangIdentifier]):
        """Initialize the class instance."""
        self.sctx = sctx
        self.default = None
        self.name = name
        self.error_tag = None
        self.error_message = None
        self.units = None

    def __contains__(self: "DataType", val: ScalarValue) -> bool:
        """Return ``True`` if the receiver type contains `val`.

        If the result is ``False``, set also `error_tag` and `error_message`
        properties.
        """
        return True

    def __str__(self: "DataType"):
        """Return YANG name of the receiver type."""
        base = self.yang_type()
        return f"{self.name}({base})" if self.name else base

    def from_raw(self: "DataType", raw: RawScalar) -> Optional[ScalarValue]:
        """Return a cooked value of the receiver type.

        The input argument should follow the rules for JSON
        representation of scalar values as specified in
        [RFC7951]_. Conformance to the receiving type isn't
        guaranteed.

        Args:
            raw: Raw value obtained from JSON parser.

        """
        if isinstance(raw, str):
            return raw

    def from_xml(self: "DataType", xml: ET.Element) -> Optional[ScalarValue]:
        """Return a cooked value of the received XML type.

        Args:
            xml: Text of the XML node
        """
        if isinstance(xml.text, str):
            return xml.text

    def to_raw(self: "DataType", val: ScalarValue) -> Optional[RawScalar]:
        """Return a raw value ready to be serialized in JSON."""
        return val

    def to_xml(self: "DataType", val: ScalarValue) -> Optional[str]:
        """Return XML text value ready to be serialized in XML."""
        return str(val)

    def parse_value(self: "DataType", text: str) -> Optional[ScalarValue]:
        """Parse value of the receiver's type.

        The input text should follow the rules for lexical
        representation of scalar values as specified in
        [RFC7950]_. Conformance to the receiving type isn't
        guaranteed.

        Args:
            text: String representation of the value.

        Returns:
            A value of the receiver's type or ``None`` if parsing fails.

        """
        return self.from_raw(text)

    def canonical_string(self: "DataType", val: ScalarValue) -> Optional[str]:
        """Return canonical form of a value."""
        return str(val)

    def from_yang(self: "DataType", text: str) -> ScalarValue:
        """Parse value specified as default in a YANG module.

        Conformance to the receiving type isn't guaranteed.

        Args:
            text: String representation of the value.

        Raises:
            InvalidArgument: If the receiver type cannot parse the text.
        """
        res = self.parse_value(text)
        if res is None:
            raise InvalidArgument(text)
        return res

    def yang_type(self: "DataType") -> YangIdentifier:
        """Return YANG name of the receiver."""
        return self.__class__.__name__[:-4].lower()

    def _set_error_info(self: "DataType", error_tag: str = None, error_message: str = None):
        self.error_tag = error_tag if error_tag else "invalid-type"
        self.error_message = (error_message if error_message else
                              "expected " + str(self))

    def _post_process(self: "DataType", tnode: "TerminalNode") -> None:
        """Post-process the receiver type on behalf of a terminal node.

        By default do nothing.
        """
        pass

    @classmethod
    def _resolve_type(cls, stmt: Statement, sctx: SchemaContext) -> "DataType":
        typ = stmt.argument
        if typ in cls.dtypes:
            res = cls.dtypes[typ](sctx, None)
            res._handle_properties(stmt, sctx)
        else:
            p, s, loc = typ.partition(":")
            res = cls._derived_type(stmt, sctx, loc if s else typ)
        return res

    @classmethod
    def _derived_type(cls, stmt: Statement, sctx: SchemaContext,
                      name: YangIdentifier) -> "DataType":
        tchain = []
        ts = stmt
        sc = sctx
        while True:
            tdef, sc = sctx.schema_data.get_definition(ts, sc)
            ts = tdef.find1("type", required=True)
            tchain.append((tdef, ts, sc))
            if ts.argument in cls.dtypes:
                break
        res = cls.dtypes[ts.argument](sctx, name)
        btyp = True
        while tchain:
            tdef, typst, tsc = tchain.pop()
            if btyp:
                res._handle_properties(ts, sc)
                btyp = False
            else:
                res._handle_restrictions(typst, tsc)
            ust = tdef.find1("units")
            if ust:
                res.units = ust.argument
            dfst = tdef.find1("default")
            if dfst:
                res.default = res.from_yang(dfst.argument)
                if res.default is None:
                    raise InvalidArgument(dfst.argument)
        res._handle_restrictions(stmt, sctx)
        return res

    def _handle_properties(self: "DataType", stmt: Statement, sctx: SchemaContext) -> None:
        """Handle type substatements."""
        self._handle_restrictions(stmt, sctx)

    def _handle_restrictions(self: "DataType", stmt: Statement, sctx: SchemaContext) -> None:
        """Handle type restriction substatements."""
        pass

    def _type_digest(self: "DataType", config: bool) -> dict[str, Any]:
        """Return receiver's type digest.

        Args:
            config: Specifies whether the type is on a configuration node.
        """
        res = {"base": self.yang_type()}
        if self.name is not None:
            res["derived"] = self.name
        return res


class EmptyType(DataType):
    """Class representing YANG "empty" type."""

    def canonical_string(self: "EmptyType", val: tuple[None]) -> Optional[str]:
        return ""

    def __contains__(self: "EmptyType", val: tuple[None]) -> bool:
        if val == (None,):
            return True
        self._set_error_info()
        return False

    def parse_value(self: "EmptyType", text: str) -> Optional[tuple[None]]:
        if text == "":
            return (None,)

    def from_raw(self: "EmptyType", raw: RawScalar) -> Optional[tuple[None]]:
        if raw == [None]:
            return (None,)

    def to_raw(self: "EmptyType", val: tuple[None]) -> list[None]:
        return [None]

    def from_xml(self: "EmptyType", xml: ET.Element) -> Optional[tuple[None]]:
        if xml.text == None:
            return (None,)

    def to_xml(self: "EmptyType", val: tuple[None]) -> None:
        return None


class BitsType(DataType):
    """Class representing YANG "bits" type."""

    def __init__(self: "BitsType", sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.bit = {}

    def sorted_bits(self: "BitsType") -> list[tuple[str, int]]:
        """Return list of bit items sorted by position."""
        return sorted(self.bit.items(), key=lambda x: x[1])

    def from_raw(self: "BitsType", raw: RawScalar) -> Optional[tuple[str]]:
        try:
            return tuple(raw.split())
        except AttributeError:
            return None

    def from_xml(self: "BitsType", xml: ET.Element) -> Optional[tuple[str]]:
        try:
            return tuple(xml.text.split())
        except AttributeError:
            return None

    def __contains__(self: "BitsType", val: tuple[str]) -> bool:
        for b in val:
            if b not in self.bit:
                self._set_error_info(error_message="unknown bit " + b)
                return False
        return True

    def to_raw(self: "BitsType", val: tuple[str]) -> str:
        return self.canonical_string(val)

    def to_xml(self: "BitsType", val: tuple[str]) -> str:
        return self.canonical_string(val)

    def as_int(self: "BitsType", val: tuple[str]) -> int:
        """Transform a "bits" value to an integer."""
        res = 0
        try:
            for b in val:
                res += 1 << self.bit[b]
        except KeyError:
            return None
        return res

    def canonical_string(self: "BitsType", val: tuple[str]) -> Optional[str]:
        try:
            items = [(self.bit[b], b) for b in val]
        except KeyError:
            return None
        items.sort()
        return " ".join([x[1] for x in items])

    def _handle_properties(self: "BitsType", stmt: Statement, sctx: SchemaContext) -> None:
        """Handle **bit** statements."""
        nextpos = 0
        for bst in stmt.find_all("bit"):
            if not sctx.schema_data.if_features(bst, sctx.text_mid):
                continue
            label = bst.argument
            pst = bst.find1("position")
            if pst:
                pos = int(pst.argument)
                self.bit[label] = pos
                if pos > nextpos:
                    nextpos = pos
            else:
                self.bit[label] = nextpos
            nextpos += 1

    def _handle_restrictions(self: "BitsType", stmt: Statement, sctx: SchemaContext) -> None:
        bst = stmt.find_all("bit")
        if not bst:
            return
        new = set([b.argument for b in bst if
                   sctx.schema_data.if_features(b, sctx.text_mid)])
        for bit in set(self.bit) - new:
            del self.bit[bit]

    def _type_digest(self: "BitsType", config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        bits = []
        i = 0
        for b in self.sorted_bits():
            while i < b[1]:
                bits.append(None)
                i += 1
            bits.append(b[0])
        res["bits"] = bits
        return res


class BooleanType(DataType):
    """Class representing YANG "boolean" type."""

    def __contains__(self: "BooleanType", val: bool) -> bool:
        if isinstance(val, bool):
            return True
        self._set_error_info()
        return False

    def from_raw(self: "BooleanType", raw: RawScalar) -> Optional[bool]:
        """Override superclass method."""
        if isinstance(raw, bool):
            return raw

    def from_xml(self: "BooleanType", xml: ET.Element) -> Optional[ScalarValue]:
        """Return a cooked value of the received XML type.

        Args:
            xml: Text of the XML node
        """
        return self.parse_value(xml.text)

    def parse_value(self: "BooleanType", text: str) -> Optional[bool]:
        """Parse boolean value.

        Args:
            text: String representation of the value.
        """
        if text == "true":
            return True
        if text == "false":
            return False

    def canonical_string(self: "BooleanType", val: bool) -> Optional[str]:
        if val is True:
            return "true"
        if val is False:
            return "false"

    def to_xml(self: "BooleanType", val: bool) -> str:
        return self.canonical_string(val)


class LinearType(DataType):
    """Abstract class representing character or byte sequences."""

    def __init__(self: "LinearType", sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.length = None  # type: Optional[Intervals]

    def _handle_restrictions(self: "LinearType", stmt: Statement, sctx: SchemaContext) -> None:
        lstmt = stmt.find1("length")
        if lstmt:
            if self.length is None:
                self.length = Intervals(
                    [[0, 4294967295]], error_message="invalid length")
            self.length.restrict_with(lstmt.argument, *lstmt.get_error_info())

    def __contains__(self: "LinearType", val: Union[str, bytes]) -> bool:
        if self.length and len(val) not in self.length:
            self._set_error_info(self.length.error_tag, self.length.error_message)
            return False
        return True

    def _type_digest(self: "LinearType", config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        if self.length:
            res["length"] = self.length.intervals
        return res


class StringType(LinearType):
    """Class representing YANG "string" type."""

    def __init__(self: "StringType", sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.patterns = []  # type: list[Pattern]

    def _handle_restrictions(self: "StringType", stmt: Statement, sctx: SchemaContext) -> None:
        super()._handle_restrictions(stmt, sctx)
        for pst in stmt.find_all("pattern"):
            invm = pst.find1("modifier", "invert-match") is not None
            self.patterns.append(Pattern(
                pst.argument, invm, *pst.get_error_info()))

    def __contains__(self: "StringType", val: str) -> bool:
        if not isinstance(val, str):
            self._set_error_info()
            return False
        if not super().__contains__(val):
            return False
        for p in self.patterns:
            if (p.regex.match(val) is not None) == p.invert_match:
                self._set_error_info(p.error_tag, p.error_message + ': ' + val)
                return False
        return True

    def _type_digest(self: "StringType", config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        pats = [p.pattern for p in self.patterns if not p.invert_match]
        ipats = [p.pattern for p in self.patterns if p.invert_match]
        if pats:
            res["patterns"] = pats
        if ipats:
            res["neg_patterns"] = ipats
        return res


class BinaryType(LinearType):
    """Class representing YANG "binary" type."""

    def from_raw(self: "BinaryType", raw: RawScalar) -> Optional[bytes]:
        """Override superclass method."""
        try:
            return base64.b64decode(raw, validate=True)
        except (TypeError, binascii.Error):
            return None

    def from_xml(self: "BinaryType", xml: ET.Element) -> Optional[bytes]:
        """Override superclass method."""
        try:
            return base64.b64decode(xml.text, validate=True)
        except TypeError:
            return None

    def __contains__(self: "BinaryType", val: bytes) -> bool:
        if not isinstance(val, bytes):
            self._set_error_info()
            return False
        return super().__contains__(val)

    def to_raw(self: "BinaryType", val: bytes) -> str:
        return self.canonical_string(val)

    def to_xml(self: "BinaryType", val: bytes) -> str:
        return self.canonical_string(val)

    def canonical_string(self: "BinaryType", val: bytes) -> Optional[str]:
        return base64.b64encode(val).decode("ascii")


class EnumerationType(DataType):
    """Class representing YANG "enumeration" type."""

    def __init__(self: "EnumerationType", sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.enum = {}  # type: dict[str, int]

    def sorted_enums(self: "EnumerationType") -> list[tuple[str, int]]:
        """Return list of enum items sorted by value."""
        return sorted(self.enum.items(), key=lambda x: x[1])

    def __contains__(self: "EnumerationType", val: str) -> bool:
        if val in self.enum:
            return True
        self._set_error_info()
        return False

    def _handle_properties(self: "EnumerationType", stmt: Statement, sctx: SchemaContext) -> None:
        """Handle **enum** statements."""
        nextval = 0
        for est in stmt.find_all("enum"):
            if not sctx.schema_data.if_features(est, sctx.text_mid):
                continue
            label = est.argument
            vst = est.find1("value")
            if vst:
                val = int(vst.argument)
                self.enum[label] = val
                if val > nextval:
                    nextval = val
            else:
                self.enum[label] = nextval
            nextval += 1

    def _handle_restrictions(self: "EnumerationType", stmt: Statement, sctx: SchemaContext) -> None:
        est = stmt.find_all("enum")
        if not est:
            return
        new = set([e.argument for e in est if
                   sctx.schema_data.if_features(e, sctx.text_mid)])
        for en in set(self.enum) - new:
            del self.enum[en]

    def _type_digest(self: "EnumerationType", config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        if config:
            res["enums"] = list(self.enum.keys())
        return res


class LinkType(DataType):
    """Abstract class for instance-referencing types."""

    def __init__(self: "LinkType", sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.require_instance = True  # type: bool

    def _handle_restrictions(self: "LinkType", stmt: Statement, sctx: SchemaContext) -> None:
        if stmt.find1("require-instance", "false"):
            self.require_instance = False


class LeafrefType(LinkType):
    """Class representing YANG "leafref" type."""

    def __init__(self: "LeafrefType", sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.path = None
        self.ref_type = None

    def _handle_properties(self: "LeafrefType", stmt: Statement, sctx: SchemaContext) -> None:
        super()._handle_properties(stmt, sctx)
        self.path = XPathParser(
            stmt.find1("path", required=True).argument, sctx).parse()

    def canonical_string(self: "LeafrefType", val: ScalarValue) -> Optional[str]:
        return self.ref_type.canonical_string(val)

    def __contains__(self: "LeafrefType", val: ScalarValue) -> bool:
        return val in self.ref_type

    def from_raw(self: "LeafrefType", raw: RawScalar) -> Optional[ScalarValue]:
        return self.ref_type.from_raw(raw)

    def from_xml(self: "LeafrefType", xml: ET.Element) -> Optional[bytes]:
        return self.ref_type.from_xml(xml)

    def to_raw(self: "LeafrefType", val: ScalarValue) -> RawScalar:
        return self.ref_type.to_raw(val)

    def to_xml(self: "LeafrefType", val: ScalarValue) -> RawScalar:
        return self.ref_type.to_xml(val)

    def parse_value(self: "LeafrefType", text: str) -> Optional[ScalarValue]:
        return self.ref_type.parse_value(text)

    def from_yang(self: "LeafrefType", text: str) -> ScalarValue:
        return self.ref_type.from_yang(text)

    def _deref(self: "LeafrefType", node: InstanceNode) -> list[InstanceNode]:
        ns = self.path.evaluate(node)
        return [n for n in ns if str(n) == str(node)]

    def _post_process(self: "LeafrefType", tnode: "TerminalNode") -> None:
        ref = tnode._follow_leafref(self.path, tnode)
        if ref is None:
            raise InvalidLeafrefPath(tnode.qual_name)
        self.ref_type = ref.type

    def _type_digest(self: "LeafrefType", config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        res["ref_type"] = self.ref_type._type_digest(config)
        return res


class InstanceIdentifierType(LinkType):
    """Class representing YANG "instance-identifier" type."""

    def __str__(self: "InstanceIdentifierType"):
        return "instance-identifier"

    def yang_type(self: "InstanceIdentifierType") -> YangIdentifier:
        """Override the superclass method."""
        return "instance-identifier"

    def from_raw(self: "InstanceIdentifierType", raw: RawScalar) -> Optional[InstanceRoute]:
        try:
            return InstanceIdParser(raw).parse()
        except ParserException:
            return None

    def from_xml(self: "InstanceIdentifierType", xml: ET.Element) -> Optional[InstanceRoute]:
        return self.from_raw(xml.text)

    def to_raw(self: "InstanceIdentifierType", val: InstanceRoute) -> str:
        """Override the superclass method."""
        return str(val)

    def to_xml(self: "InstanceIdentifierType", val: InstanceRoute) -> str:
        """Override the superclass method."""
        return str(val)

    def from_yang(self: "InstanceIdentifierType", text: str) -> InstanceRoute:
        """Override the superclass method."""
        return XPathParser(text, self.sctx).parse().as_instance_route()

    @staticmethod
    def _deref(node: InstanceNode) -> list[InstanceNode]:
        return [node.top().goto(node.value)]


class IdentityrefType(DataType):
    """Class representing YANG "identityref" type."""

    def __init__(self: "IdentityrefType", sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.bases = []  # type: list[QualName]

    def from_raw(self: "IdentityrefType", raw: RawScalar) -> Optional[QualName]:
        try:
            i1, s, i2 = raw.partition(":")
        except AttributeError:
            return None
        return (i2, i1) if s else (i1, self.sctx.default_ns)

    def from_xml(self: "IdentityrefType", xml: ET.Element) -> Optional[QualName]:
        try:
            i1, s, i2 = xml.text.partition(":")
        except AttributeError:
            return None

        if not s:
            # not prefixed, so assume default ns
            return (i1, self.sctx.default_ns)
        elif i1 == self.sctx.default_ns:
            # prefixed with default ns
            return (i2, self.sctx.default_ns)

        # the following code has issues (Issue #79)
        ns_url = xml.attrib.get('xmlns:'+i1)
        if not ns_url:
            raise MissingModuleNamespace(ns_url)
        module = self.sctx.schema_data.modules_by_ns.get(ns_url)
        if not module:
            raise MissingModuleNamespace(ns_url)
        return (i2, module.main_module[0])

    def __contains__(self: "IdentityrefType", val: QualName) -> bool:
        for b in self.bases:
            if not self.sctx.schema_data.is_derived_from(val, b):
                self._set_error_info(
                    error_message=f"'{self.canonical_string(val)}'" +
                    f" not derived from '{self.canonical_string(b)}'")
                return False
        return True

    def to_raw(self: "IdentityrefType", val: QualName) -> str:
        return self.canonical_string(val)

    def to_xml(self: "IdentityrefType", val: QualName) -> str:
        return self.canonical_string(val)

    def from_yang(self: "IdentityrefType", text: str) -> QualName:
        """Override the superclass method."""
        try:
            return self.sctx.schema_data.translate_pname(
                text, self.sctx.text_mid)
        except (ModuleNotRegistered, UnknownPrefix):
            raise InvalidArgument(text)

    def canonical_string(self: "IdentityrefType",
                         val: ScalarValue) -> Optional[str]:
        """Return canonical form of a value."""
        return f"{val[1]}:{val[0]}"

    def _handle_properties(self: "IdentityrefType", stmt: Statement, sctx: SchemaContext) -> None:
        self.bases = []
        for b in stmt.find_all("base"):
            self.bases.append(
                sctx.schema_data.translate_pname(b.argument, sctx.text_mid))

    def _type_digest(self: "IdentityrefType", config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        if config:
            res["identities"] = list(
                self.sctx.schema_data.derived_from_all(self.bases))
        return res


class NumericType(DataType):
    """Abstract class for numeric data types."""

    def __init__(self: "NumericType", sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.range = None  # type: Optional[Intervals]

    def __contains__(self: "NumericType", val: Union[int, decimal.Decimal]) -> bool:
        if self.range is None:
            if self._range[0] <= val <= self._range[1]:
                return True
            else:
                self._set_error_info()
                return False
        if val in self.range:
            return True
        self._set_error_info(self.range.error_tag, self.range.error_message)
        return False

    def _handle_restrictions(self: "NumericType", stmt: Statement, sctx: SchemaContext) -> None:
        rstmt = stmt.find1("range")
        if rstmt:
            if self.range is None:
                self.range = Intervals([self._range], parser=self.parse_value,
                                       error_message="not in range")
            self.range.restrict_with(rstmt.argument, *rstmt.get_error_info())

    def _type_digest(self: "NumericType", config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        if self.range:
            res["range"] = [[self.to_raw(r[0]), self.to_raw(r[-1])]
                            for r in self.range.intervals]
        return res


class Decimal64Type(NumericType):
    """Class representing YANG "decimal64" type."""

    def __init__(self: "Decimal64Type", sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self._epsilon = decimal.Decimal(0)  # type: decimal.Decimal

    @property
    def _range(self: "Decimal64Type") -> list[decimal.Decimal]:
        quot = decimal.Decimal(10**self.fraction_digits)
        lim = decimal.Decimal(9223372036854775808)
        return [-lim / quot, (lim - 1) / quot]

    def _handle_properties(self: "Decimal64Type", stmt: Statement, sctx: SchemaContext) -> None:
        self.fraction_digits = int(
            stmt.find1("fraction-digits", required=True).argument)
        self._epsilon = decimal.Decimal(10) ** -self.fraction_digits
        super()._handle_properties(stmt, sctx)

    def from_raw(self: "Decimal64Type", raw: RawScalar) -> Optional[decimal.Decimal]:
        """Override superclass method.

        According to [RFC7951]_, a raw instance must be string.
        """
        if not isinstance(raw, str):
            return None
        try:
            return decimal.Decimal(raw).quantize(self._epsilon)
        except decimal.InvalidOperation:
            return None

    def from_xml(self: "Decimal64Type", xml: ET.Element) -> Optional[decimal.Decimal]:
        try:
            return decimal.Decimal(xml.text).quantize(self._epsilon)
        except decimal.InvalidOperation:
            return None

    def to_raw(self: "Decimal64Type", val: decimal.Decimal) -> str:
        return self.canonical_string(val)

    def to_xml(self: "Decimal64Type", val: decimal.Decimal) -> str:
        return self.canonical_string(val)

    def canonical_string(self: "Decimal64Type", val: decimal.Decimal) -> Optional[str]:
        if val == 0:
            return "0.0"
        sval = str(val.quantize(self._epsilon)).rstrip("0")
        return (sval + "0") if sval.endswith(".") else sval

    def __contains__(self: "Decimal64Type", val: decimal.Decimal) -> bool:
        if not isinstance(val, decimal.Decimal):
            self._set_error_info()
            return False
        return super().__contains__(val)

    def _type_digest(self: "Decimal64Type", config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        res["fraction_digits"] = self.fraction_digits
        return res


class IntegralType(NumericType):
    """Abstract class for integral data types."""

    octhex = re.compile("[-+]?0([x0-9])")
    """Regular expression for octal or hexadecimal default."""

    def __contains__(self: "IntegralType", val: int) -> bool:
        if not isinstance(val, int) or isinstance(val, bool):
            self._set_error_info()
            return False
        return super().__contains__(val)

    def parse_value(self: "IntegralType", text: str) -> Optional[int]:
        """Override superclass method."""
        try:
            return int(text)
        except ValueError:
            return None

    def from_raw(self: "IntegralType", raw: RawScalar) -> Optional[int]:
        if not isinstance(raw, int) or isinstance(raw, bool):
            return None
        try:
            return int(raw)
        except (ValueError, TypeError):
            return None

    def from_xml(self: "IntegralType", xml: ET.Element) -> Optional[int]:
        try:
            return int(xml.text)
        except (ValueError, TypeError):
            return None

    def from_yang(self: "IntegralType", text: str) -> int:
        """Override the superclass method."""
        mo = self.octhex.match(text)
        if mo:
            base = 16 if mo.group(1) == "x" else 8
        else:
            base = 10
        try:
            return int(text, base)
        except (ValueError, TypeError):
            raise InvalidArgument(text)


class Int8Type(IntegralType):
    """Class representing YANG "int8" type."""

    _range = [-128, 127]


class Int16Type(IntegralType):
    """Class representing YANG "int16" type."""

    _range = [-32768, 32767]


class Int32Type(IntegralType):
    """Class representing YANG "int32" type."""

    _range = [-2147483648, 2147483647]


class Int64Type(IntegralType):
    """Class representing YANG "int64" type."""

    _range = [-9223372036854775808, 9223372036854775807]

    def from_raw(self: "Int64Type", raw: RawScalar) -> Optional[int]:
        """Override superclass method.

        According to [RFC7951]_, a raw instance must be string.
        """
        if not isinstance(raw, str) or isinstance(raw, bool):
            return None
        try:
            return int(raw)
        except (ValueError, TypeError):
            return None

    def from_xml(self: "Int64Type", xml: ET.Element) -> Optional[int]:
        """Override superclass method.

        XML is always delivered as a text element
        """
        try:
            return int(xml.text)
        except (ValueError, TypeError):
            return None

    def to_raw(self: "Int64Type", val: int) -> str:
        return self.canonical_string(val)

    def to_xml(self: "Int64Type", val: int) -> str:
        return self.canonical_string(val)


class Uint8Type(IntegralType):
    """Class representing YANG "uint8" type."""

    _range = [0, 255]


class Uint16Type(IntegralType):
    """Class representing YANG "uint16" type."""

    _range = [0, 65535]


class Uint32Type(IntegralType):
    """Class representing YANG "uint32" type."""

    _range = [0, 4294967295]


class Uint64Type(IntegralType):
    """Class representing YANG "uint64" type."""

    _range = [0, 18446744073709551615]

    def from_raw(self: "Uint64Type", raw: str) -> Optional[int]:
        """Override superclass method.

        According to [RFC7951]_, a raw instance must be string.
        """
        if not isinstance(raw, str) or isinstance(raw, bool):
            return None
        try:
            return int(raw)
        except (ValueError, TypeError):
            return None

    def from_xml(self: "Uint64Type", xml: ET.Element) -> Optional[int]:
        """Override superclass method.

        XML is always delivered as a text element
        """
        try:
            return int(xml.text)
        except (ValueError, TypeError):
            return None

    def to_raw(self: "Uint64Type", val: int) -> str:
        return self.canonical_string(val)

    def to_xml(self: "Uint64Type", val: int) -> str:
        return self.canonical_string(val)


class UnionType(DataType):
    """Class representing YANG "union" type."""

    def __init__(self: "UnionType", sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.types = []  # type: list[DataType]

    def to_raw(self: "UnionType", val: ScalarValue) -> RawScalar:
        for t in self.types:
            if val in t:
                return t.to_raw(val)

    def to_xml(self: "UnionType", val: ScalarValue) -> RawScalar:
        for t in self.types:
            if val in t:
                return t.to_xml(val)

    def canonical_string(self: "UnionType", val: ScalarValue) -> Optional[str]:
        for t in self.types:
            if val in t:
                return t.canonical_string(val)
        return None

    def parse_value(self: "UnionType", text: str) -> Optional[ScalarValue]:
        for t in self.types:
            val = t.parse_value(text)
            if val is not None and val in t:
                return val
        return None

    def from_raw(self: "UnionType", raw: RawScalar) -> Optional[ScalarValue]:
        for t in self.types:
            val = t.from_raw(raw)
            if val is not None and val in t:
                return val
        return None

    def from_xml(self: "UnionType", xml: ET.Element) -> Optional[ScalarValue]:
        for t in self.types:
            val = t.from_xml(xml)
            if val is not None and val in t:
                return val
        return None

    def __contains__(self: "UnionType", val: Any) -> bool:
        for t in self.types:
            try:
                if val in t:
                    return True
            except TypeError:
                continue
        return False

    def _handle_properties(self: "UnionType", stmt: Statement, sctx: SchemaContext) -> None:
        self.types = [self._resolve_type(ts, sctx)
                      for ts in stmt.find_all("type")]

    def _post_process(self: "UnionType", tnode: "TerminalNode") -> None:
        for t in self.types:
            t._post_process(tnode)


DataType.dtypes = {"binary": BinaryType,
                   "bits": BitsType,
                   "boolean": BooleanType,
                   "decimal64": Decimal64Type,
                   "empty": EmptyType,
                   "enumeration": EnumerationType,
                   "identityref": IdentityrefType,
                   "instance-identifier": InstanceIdentifierType,
                   "int8": Int8Type,
                   "int16": Int16Type,
                   "int32": Int32Type,
                   "int64": Int64Type,
                   "leafref": LeafrefType,
                   "string": StringType,
                   "uint8": Uint8Type,
                   "uint16": Uint16Type,
                   "uint32": Uint32Type,
                   "uint64": Uint64Type,
                   "union": UnionType
                   }
"""Dictionary mapping type names to classes."""



# Copyright © 2016-2023 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Conversion of YANG library formats.

This module defines an entry point for converting RFC-8525-style YANG
library data to the old RFC 7895 format that is used by Yangson.
"""

import argparse
import json
import os
import sys
from yangson import DataModel
from yangson.enumerations import ContentType
from yangson.exceptions import (
    NonexistentInstance, ValidationError, RawDataError)
from yangson.instance import ArrayEntry
from yangson.typealiases import RawObject

YL7895 = """
{
  "ietf-yang-library:modules-state": {
    "module-set-id": "",
    "module": [
      {
        "name": "ietf-yang-library",
        "revision": "2016-06-21",
        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-library",
        "conformance-type": "implement"
      },
      {
        "name": "ietf-inet-types",
        "revision": "2013-07-15",
        "namespace": "urn:ietf:params:xml:ns:yang:ietf-inet-types",
        "conformance-type": "import"
      },
      {
        "name": "ietf-yang-types",
        "revision": "2013-07-15",
        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-types",
        "conformance-type": "import"
      }
    ]
  }
}
"""

YL8525 = """
{
  "ietf-yang-library:modules-state": {
    "module-set-id": "",
    "module": [
      {
        "name": "ietf-yang-library",
        "revision": "2019-01-04",
        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-library",
        "conformance-type": "implement"
      },
      {
        "name": "ietf-inet-types",
        "revision": "2013-07-15",
        "namespace": "urn:ietf:params:xml:ns:yang:ietf-inet-types",
        "conformance-type": "import"
      },
      {
        "name": "ietf-datastores",
        "revision": "2018-02-14",
        "namespace": "urn:ietf:params:xml:ns:yang:ietf-datastores",
        "conformance-type": "implement"
      },
      {
        "name": "ietf-yang-types",
        "revision": "2013-07-15",
        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-types",
        "conformance-type": "import"
      }
    ]
  }
}
"""


class ModuleData:
    name: str
    revision: str
    location: set[str]

    def __init__(self: "ModuleData", rfc8525_entry: ArrayEntry):
        """Initialize the receiver."""
        val = rfc8525_entry.value
        self.name = val["name"]
        self.revision = val.get("revision", "")
        self.location = set([loc.strip() for loc in
                             val.get("location", [])])

    def key(self) -> tuple[str, str]:
        """Return the receiver's key (module name & revision)."""
        return (self.name, self.revision)

    def merge(self, other: "ModuleData") -> None:
        """Merge the receiver with another instance."""
        self.location |= other.location

    def as_raw(self) -> RawObject:
        """Return the receiver represented as an RFC 7895 entry."""
        res = {
            "name": self.name,
            "revision": self.revision
        }
        if self.location:
            res["schema"] = list(self.location)[0]
        return res


class MainModuleData(ModuleData):
    namespace: str
    import_only: bool
    deviation: set[str]
    feature: set[str]
    submodule: dict[tuple[str, str], ModuleData]

    def __init__(self: "MainModuleData", rfc8525_entry: ArrayEntry,
                 import_only: bool) -> None:
        """Initialize the receiver."""
        super().__init__(rfc8525_entry)
        val = rfc8525_entry.value
        self.import_only = import_only
        self.namespace = val["namespace"].strip()
        self.submodule = {}
        if "submodule" in val:
            for s in rfc8525_entry["submodule"]:
                self.add_submodule(s)
        self.feature = set(val.get("feature", []))
        self.deviation = set()
        if "deviation" in val:
            for dev in rfc8525_entry["deviation"]:
                dmod = dev._deref()[0]
                try:
                    rev = dmod.sibling("revision").value
                except NonexistentInstance:
                    rev = ""
                self.deviation.add((dmod.value, rev))

    def add_submodule(self, sub_entry: ArrayEntry):
        """Add or merge submodule defined in an RFC 8525 submodule entry."""
        smod = ModuleData(sub_entry)
        key = smod.key()
        if key in self.submodule:
            self.submodule[key].merge(smod)
        else:
            self.submodule[key] = smod

    def merge(self, other: "MainModuleData") -> None:
        """Extend the superclass method."""
        super().merge(other)
        self.deviation |= other.deviation
        self.feature |= other.feature
        self.import_only = self.import_only or other.import_only
        for sm in other.submodule:
            self.add_submodule(sm)

    def as_raw(self) -> RawObject:
        """Extend the superclass method."""
        res = super().as_raw()
        res["conformance-type"] = "import" if self.import_only else "implement"
        res["namespace"] = self.namespace
        if self.submodule:
            res["submodule"] = [self.submodule[s].as_raw()
                                for s in self.submodule]
        if self.feature:
            res["feature"] = list(self.feature)
        if self.deviation:
            res["deviation"] = [{ "name": n, "revision": r }
                                for (n,r) in self.deviation]
        return res


def main() -> int:
    """Entry-point for a conversion script.
    """
    parser = argparse.ArgumentParser(
        prog="convert8525",
        description=("Convert NMDA-compatible YANG library data to"
                     " the old style of RFC 7985."))
    parser.add_argument(
        "ylib", metavar="YLIB",
        help=("name of the input file containing JSON-encoded"
                "YANG library data conforming to the RFC 8525 schema"))
    parser.add_argument(
        "-o", "--output", metavar="OUTFILE",
        help="Direct output to OUTFILE instead of standard output")
    parser.add_argument(
        "-p", "--path",
        help=("colon-separated list of directories to search"
              " for YANG modules"))
    xgroup = parser.add_mutually_exclusive_group()
    xgroup.add_argument(
        "-d", "--datastore", metavar="DSTORE",
        default="running",
        help="name of a datastore")
    xgroup.add_argument(
        "-s", "--schema", metavar="SCHEMA",
        help="name of a schema")
    args = parser.parse_args()
    fn8525: str = args.ylib
    fn7895: Optional[str] = args.output
    path: Optional[str] = args.path
    datastore: str = args.datastore
    schema: Optional[str] = args.schema
    sp = path if path else os.environ.get("YANG_MODPATH", ".")
    dm8525 = DataModel(YL8525, sp.split(":"))
    try:
        with open(fn8525) as infile:
            ri = json.load(infile)
        inst = dm8525.from_raw(ri)
        inst.validate(ctype=ContentType.nonconfig)
    except (FileNotFoundError, PermissionError, json.decoder.JSONDecodeError,
            RawDataError, ValidationError) as e:
        print ("Invalid input data:", str(e), file=sys.stderr)
        return 1
    top = inst["ietf-yang-library:yang-library"]
    if schema is None:
        try:
            schema = top["datastore"].look_up(
                name=(datastore, "ietf-datastores"))["schema"].value
        except NonexistentInstance:
            print(f"No such datastore: 'ietf-datastores:{datastore}'",
                  file=sys.stderr)
            return 2
    try:
        msets = top["schema"].look_up(name=schema)["module-set"]
    except NonexistentInstance:
        print(f"No such schema: '{schema}'", file=sys.stderr)
        return 2
    modules: dict[tuple[str, str], MainModuleData] = {}
    for ms in msets:
        msentry = top["module-set"].look_up(name=ms.value)
        if "module" in msentry:
            for yam in msentry["module"]:
                ment = MainModuleData(yam, import_only = False)
                modrev = ment.key()
                if modrev in modules:
                    modules[modrev].merge(ment)
                else:
                    modules[modrev] = ment
        if "import-only-module" in msentry:
            for yam in msentry["import-only-module"]:
                ment = MainModuleData(yam, import_only = True)
                modrev = ment.key()
                if modrev in modules:
                    modules[modrev].merge(ment)
                else:
                    modules[modrev] = ment
    dm7895 = DataModel(YL7895, sp.split(":"))
    res = dm7895.from_raw(
        { "ietf-yang-library:modules-state":
          { "module-set-id": top["content-id"].value }})
    rtop = res["ietf-yang-library:modules-state"]
    res = rtop.put_member("module",
                    [modules[m].as_raw() for m in modules], raw = True).up().up()
    res.validate(ctype=ContentType.nonconfig)
    try:
        outf = open(args.output, mode="w") if args.output else sys.stdout
    except FileNotFoundError as e:
        print(e, file=sys.stderr)
        return 3
    json.dump(res.raw_value(), outf, indent=2)
    outf.close()
    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright © 2016–2023 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""YANG statements and a parser for YANG modules.

This module implements the following classes:

* ModuleParser: Recursive-descent parser for YANG modules.
* Statement: YANG statements.
"""
from typing import Optional
from .exceptions import (
    EndOfInput, StatementNotFound, UnexpectedInput, InvalidArgument,
    ModuleNameMismatch, ModuleRevisionMismatch)
from .parser import Parser
from .typealiases import YangIdentifier


class Statement:

    """YANG statement."""

    _escape_table = str.maketrans({'"': '\\"', '\\': '\\\\'})
    """Table for translating characters to their escaped form."""

    def __init__(self: "Statement",
                 kw: YangIdentifier,
                 arg: Optional[str],
                 pref: YangIdentifier = None):
        """Initialize the class instance.

        Args:
            kw: Keyword.
            arg: Argument.
            sup: Parent statement.
            sub: List of substatements.
            pref: Keyword prefix (``None`` for built-in statements).
        """
        self.prefix = pref
        self.keyword = kw
        self.argument = arg
        self.superstmt = None
        self.substatements = []

    def __str__(self: "Statement") -> str:
        """Return string representation of the receiver.
        """
        kw = (self.keyword if self.prefix is None
              else self.prefix + ":" + self.keyword)
        arg = ("" if self.argument is None
               else ' "' + self.argument.translate(self._escape_table) + '"')
        rest = " { ... }" if self.substatements else ";"
        return kw + arg + rest

    def find1(self: "Statement", kw: YangIdentifier, arg: str = None,
              pref: YangIdentifier = None,
              required: bool = False) -> Optional["Statement"]:
        """Return first substatement with the given parameters.

        Args:
            kw: Statement keyword (local part for extensions).
            arg: Argument (all arguments will match if ``None``).
            pref: Keyword prefix (``None`` for built-in statements).
            required: Should an exception be raised on failure?

        Raises:
            StatementNotFound: If `required` is ``True`` and the
                statement is not found.
        """
        for sub in self.substatements:
            if (sub.keyword == kw and sub.prefix == pref and
                    (arg is None or sub.argument == arg)):
                return sub
        if required:
            raise StatementNotFound(str(self), kw)

    def find_all(self: "Statement", kw: YangIdentifier,
                 pref: YangIdentifier = None) -> list["Statement"]:
        """Return the list all substatements with the given keyword and prefix.

        Args:
            kw: Statement keyword (local part for extensions).
            pref: Keyword prefix (``None`` for built-in statements).
        """
        return [c for c in self.substatements
                if c.keyword == kw and c.prefix == pref]

    def _get_scoped_definition(self: "Statement", name: YangIdentifier,
                               kw: YangIdentifier) -> Optional["Statement"]:
        """Search ancestor statements for a scoped (non-importable) definition.

        Args:
            name: Name of a grouping or datatype (with no prefix).
            kw: ``grouping`` or ``typedef``.
        """
        stmt = self.superstmt
        while stmt.superstmt:
            res = stmt.find1(kw, name)
            if res:
                return res
            stmt = stmt.superstmt
        return None

    def get_error_info(self: "Statement") -> tuple[Optional[str], Optional[str]]:
        """Return receiver's error tag and error message if present."""
        etag = self.find1("error-app-tag")
        emsg = self.find1("error-message")
        return (etag.argument if etag else None, emsg.argument if emsg else None)


class ModuleParser(Parser):
    """Parse YANG modules."""

    unescape_map = {"n": "\n", "t": "\t", '"': '"',
                    "\\": "\\"}  # type: dict[str,str]
    """Dictionary for mapping escape sequences to characters."""

    def __init__(self: "ModuleParser", text: str, name: YangIdentifier = None, rev: str = None):
        """Initialize the parser instance.

        Args:
            name: Expected module name.
            rev: Expected revision date.
        """
        super().__init__(text)
        self.name = name
        self.rev = rev

    def parse(self: "ModuleParser") -> Statement:
        """Parse a complete YANG module or submodule.

        Args:
            mtext: YANG module text.

        Raises:
            EndOfInput: If past the end of input.
            ModuleNameMismatch: If parsed module name doesn't match `self.name`.
            ModuleRevisionMismatch: If parsed revision date doesn't match `self.rev`.
            UnexpectedInput: If top-level statement isn't ``(sub)module``.
        """
        self.opt_separator()
        start = self.offset
        res = self.statement()
        if res.keyword not in ["module", "submodule"]:
            self.offset = start
            raise UnexpectedInput(self, "'module' or 'submodule'")
        if self.name is not None and res.argument != self.name:
            raise ModuleNameMismatch(res.argument, self.name)
        if self.rev:
            revst = res.find1("revision")
            if revst is None or revst.argument != self.rev:
                raise ModuleRevisionMismatch(revst.argument, self.rev)
        try:
            self.opt_separator()
        except EndOfInput:
            return res
        raise UnexpectedInput(self, "end of input")

    def _back_break(self: "ModuleParser") -> int:
        self.offset -= 1
        return -1

    @classmethod
    def unescape(cls, text: str) -> str:
        """Replace escape sequence with corresponding characters.

        Args:
            text: Text to unescape.
        """
        chop = text.split("\\", 1)
        try:
            return (chop[0] if len(chop) == 1
                    else chop[0] + cls.unescape_map[chop[1][0]] +
                    cls.unescape(chop[1][1:]))
        except KeyError:
            raise InvalidArgument(text) from None

    def opt_separator(self: "ModuleParser") -> bool:
        """Parse an optional separator and return ``True`` if found.

        Raises:
            EndOfInput: If past the end of input.
        """
        start = self.offset
        self.dfa([
            {  # state 0: whitespace
                "": lambda: -1,
                " ": lambda: 0,
                "\t": lambda: 0,
                "\n": lambda: 0,
                "\r": lambda: 1,
                "/": lambda: 2
            },
            {  # state 1: CR/LF?
                "": self._back_break,
                "\n": lambda: 0
            },
            {  # state 2: start comment?
                "": self._back_break,
                "/": lambda: 3,
                "*": lambda: 4
            },
            {  # state 3: line comment
                "": lambda: 3,
                "\n": lambda: 0
            },
            {  # state 4: block comment
                "": lambda: 4,
                "*": lambda: 5
            },
            {  # state 5: end block comment?
                "": lambda: 4,
                "/": lambda: 0,
                "*": lambda: 5
            }])
        return start < self.offset

    def separator(self: "ModuleParser") -> None:
        """Parse a mandatory separator.

        Raises:
            EndOfInput: If past the end of input.
            UnexpectedInput: If no separator is found.
        """
        present = self.opt_separator()
        if not present:
            raise UnexpectedInput(self, "separator")

    def keyword(self: "ModuleParser") -> tuple[Optional[str], str]:
        """Parse a YANG statement keyword.

        Raises:
            EndOfInput: If past the end of input.
            UnexpectedInput: If no syntactically correct keyword is found.
        """
        i1 = self.yang_identifier()
        if self.peek() == ":":
            self.offset += 1
            i2 = self.yang_identifier()
            return (i1, i2)
        return (None, i1)

    def statement(self: "ModuleParser") -> Statement:
        """Parse YANG statement.

        Raises:
            EndOfInput: If past the end of input.
            UnexpectedInput: If no syntactically correct statement is found.
        """
        pref, kw = self.keyword()
        pres = self.opt_separator()
        next = self.peek()
        if next == ";":
            arg = None
            sub = False  # type: bool
        elif next == "{":
            arg = None
            sub = True
        elif not pres:
            raise UnexpectedInput(self, "separator")
        else:
            self._arg = ""
            sub = self.argument()
            arg = self._arg
        self.offset += 1
        res = Statement(kw, arg, pref=pref)
        if sub:
            res.substatements = self.substatements()
            for sub in res.substatements:
                sub.superstmt = res
        return res

    def argument(self: "ModuleParser") -> bool:
        """Parse statement argument.

        Return ``True`` if the argument is followed by block of substatements.
        """
        next = self.peek()
        if next == "'":
            quoted = True
            self.sq_argument()
        elif next == '"':
            quoted = True
            self.dq_argument()
        elif self._arg == "":
            quoted = False
            self.unq_argument()
        else:
            raise UnexpectedInput(self, "single or double quote")
        self.opt_separator()
        next = self.peek()
        if next == ";":
            return False
        if next == "{":
            return True
        elif quoted and next == "+":
            self.offset += 1
            self.opt_separator()
            return self.argument()
        else:
            raise UnexpectedInput(self, "';', '{'" +
                                  (" or '+'" if quoted else ""))

    def sq_argument(self: "ModuleParser") -> str:
        """Parse single-quoted argument.

        Raises:
            EndOfInput: If past the end of input.
        """
        self.offset += 1
        self._arg += self.up_to("'")

    def dq_argument(self: "ModuleParser") -> str:
        """Parse double-quoted argument.

        Raises:
            EndOfInput: If past the end of input.
        """
        def escape():
            self._escape = True
            return 1
        self._escape = False                 # any escaped chars?
        self.offset += 1
        start = self.offset
        self.dfa([
            {  # state 0: argument
                "": lambda: 0,
                '"': lambda: -1,
                "\\": escape
            },
            {  # state 1: after escape
                "": lambda: 0
            }])
        self._arg += (self.unescape(self.input[start:self.offset])
                      if self._escape else self.input[start:self.offset])
        self.offset += 1

    def unq_argument(self: "ModuleParser") -> str:
        """Parse unquoted argument.

        Raises:
            EndOfInput: If past the end of input.
        """
        start = self.offset
        self.dfa([
            {  # state 0: argument
                "": lambda: 0,
                ";": lambda: -1,
                " ": lambda: -1,
                "\t": lambda: -1,
                "\r": lambda: -1,
                "\n": lambda: -1,
                "{": lambda: -1,
                '/': lambda: 1
            },
            {  # state 1: comment?
                "": lambda: 0,
                "/": self._back_break,
                "*": self._back_break
            }])
        self._arg = self.input[start:self.offset]

    def substatements(self: "ModuleParser") -> list[Statement]:
        """Parse substatements.

        Raises:
            EndOfInput: If past the end of input.
        """
        res = []
        self.opt_separator()
        while self.peek() != "}":
            res.append(self.statement())
            self.opt_separator()
        self.offset += 1
        return res
